/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "MBMesh.hpp"
#include "ParCALMeshVal.hpp"
#include "PNclMBUtils.hpp"
#include "PNclMDUtils.hpp"
#include <iostream>

namespace parncl{

// ParCALMeshVal Sptr func defns
ParCALMeshValSPtr::ParCALMeshValSPtr(ParCALMeshVal *valp):valp_(valp)
{
	valp_->inc_ref_cnt();
}

ParCALMeshVal *ParCALMeshValSPtr::operator->()
{
	// Access mesh val via sptr
	return valp_;
}

ParCALMeshVal &ParCALMeshValSPtr::operator*()
{
	// Copy mesh val by value
	if(valp_->is_file_val() || (valp_->get_ref_cnt() > 1)){
		ParCALMeshVal *new_val = new ParCALMeshVal(*valp_);
		return *new_val;
	}
	valp_->set_gval_cache_dirty();
	return *valp_;
}

ParCALMeshValSPtr::ParCALMeshValSPtr(const ParCALMeshValSPtr &sptr):valp_(sptr.valp_)
{
	// Copy by ref
	valp_->inc_ref_cnt();
}

ParCALMeshValSPtr& ParCALMeshValSPtr::operator=(const ParCALMeshValSPtr &rhs)
{
	if(this != &rhs){
		valp_->dec_ref_cnt();
		if(valp_->get_ref_cnt() == 0){
			delete valp_;
		}
		valp_ = rhs.valp_;
		valp_->inc_ref_cnt();
	}
	return *this;
}

ParCALMeshValSPtr::~ParCALMeshValSPtr()
{
	valp_->dec_ref_cnt();
	if(valp_->get_ref_cnt() == 0){
		delete valp_;
	}
}

// ParCALMeshVal func defns ....

std::string ParCALMeshVal::convert_to_parcal_loc(const ParCALMeshVal::MeshValLoc val_loc)
{
	switch(val_loc){
		case MESHVAL_ON_QUAD_CENTER:
				return std::string("QUAD");
		default:
				return std::string("UNIMPL");	
	}
}

ParCALMeshVal::ParCALMeshVal(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar, void *missing_val):prev_cached_lval_chunk_(parncl::PARCALVAL_NCHUNK_NONE), prev_cached_lval_buf_(NULL), ncached_lval_chunks_(0), mesh_(NULL), mb_core_(NULL), val_loc_(ParCALMeshVal::MESHVAL_LOC_UNKNOWN), pcvar_(pcvar), data_type_(data_type), missing_val_(missing_val), num_pcvar_chunks_(0), pcvar_chunk_sz_(0), num_pcvar_levs_(0), sel_(NULL), is_fvar_(is_fvar), ref_cnt_(0)
{
	cached_gval_ = std::make_pair(false, (std::vector<double> *)NULL);
    const std::vector<pargal::pcdim>& pcvar_dims = pcvar->get_dims();
	for(std::vector<pargal::pcdim>::const_iterator iter = pcvar_dims.begin(); iter != pcvar_dims.end(); ++iter){
		if((*iter).name() == std::string("lev")){
			num_pcvar_levs_ = (*iter).size();
			break;
		}
		if((*iter).name() == std::string("ilev")){
			num_pcvar_levs_ = (*iter).size();
			break;
		}
	}
}

ParCALMeshValSPtr ParCALMeshVal::create_mesh_val(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar, void *missing_val)
{
	ParCALMeshVal *mesh_val = new ParCALMeshVal(pcvar, data_type, is_fvar, missing_val);
	mesh_val->num_pcvar_chunks_ = pcvar->get_tsteps().size();
	return (ParCALMeshValSPtr(mesh_val));
}

ParCALMeshValSPtr ParCALMeshVal::create_mesh_val(MBMesh *mesh, moab::Interface *mb_core, size_t mb_chunk_sz, const std::vector<int> &val_tsteps, const std::vector<pargal::pcdim> &val_dims, size_t nlevs, nc_type data_type, const std::string &pcal_val_loc, void *missing_val, NclSelectionRecord *sel)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

	assert(mesh != NULL);
	MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle mb_set = pcal_finfo.get_file_set();
	moab::DataType dt = PNclNCMapToMB(data_type);
	std::string pcvar_name = ParCALGetUniqStr(mesh);

	for(std::vector<int>::const_iterator iter = val_tsteps.begin(); iter != val_tsteps.end(); ++iter){
		size_t tstep = *iter;
		moab::EntityType ent_type = ParCALLocToMB(pcal_val_loc);
		mb_retval = PNclMBCreateTag(mesh, pcvar_name.c_str(), tstep, nlevs, NULL, dt, ent_type);
		if(mb_retval != moab::MB_SUCCESS){
			std::string funcname("UNKNOWN");
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating MOAB tag while creating var");
			throw pargal::pargal_except(mb_core->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
		}
	}

	pargal::pcvar *pcvar = new pargal::pcvar(pcvar_name, dt,
									val_dims, pcal_finfo, val_tsteps, true, pcal_val_loc);

	ParCALMeshVal *mesh_val = new ParCALMeshVal(pcvar, data_type, false, missing_val);
	mesh_val->num_pcvar_chunks_ = pcvar->get_tsteps().size();

	NclSelectionRecord *tmp_sel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
	if(!tmp_sel){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to alloc mem for val sel");
		throw pargal::pargal_except("Unable to alloc mem for val sel", __FILE__, __LINE__, "UNKNOWN");
	}
	PNclSelCopy(tmp_sel, sel);
	mesh_val->sel_ = tmp_sel;
	// FIXME: Find val loc from dim names
	mesh_val->set_mesh_info(mesh, mb_core, mb_chunk_sz, ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
	return (ParCALMeshValSPtr(mesh_val));
}

static void set_val_tsteps(pargal::pcvar *pcvar, NclSelectionRecord *gsel, std::vector<int> &val_tsteps)
{
	// FIXME: What about 5D arrays ?
	if(gsel->n_entries == 4){
		switch(gsel->selection[0].sel_type){
			case Ncl_SUB_ALL:
				val_tsteps = pcvar->get_tsteps();
				break;
			case Ncl_SUBSCR:
				for(size_t i=gsel->selection[0].u.sub.start; i <= gsel->selection[0].u.sub.finish; i += gsel->selection[0].u.sub.stride){
					val_tsteps.push_back(i);
				}
				break;
			case Ncl_VECSUBSCR:
				for(size_t i=0; i<gsel->selection[0].u.vec.n_ind; i++){
					val_tsteps.push_back(gsel->selection[0].u.vec.ind[i]);
				}
				break;
			default:
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Invalid subscripting method for leftmost dimension");
				throw pargal::pargal_except("Unable to create val slice, invalid subscripting method used", __FILE__, __LINE__, "Unknown");
		}
	}
	else{
		val_tsteps.push_back(0);
	}
}

static void set_val_nlevs(const ParCALMeshVal &val, NclSelectionRecord *gsel, size_t &nlevs)
{
	if(gsel->n_entries >= 3){
		int lev_idx = gsel->n_entries - 3;
		switch(gsel->selection[lev_idx].sel_type){
			case Ncl_SUB_ALL:
				nlevs = val.get_num_val_levs();
				break;	
			case Ncl_SUBSCR:
				for(size_t i=gsel->selection[lev_idx].u.sub.start; i <= gsel->selection[lev_idx].u.sub.finish; i += gsel->selection[lev_idx].u.sub.stride){
					nlevs++;
				}
				break;
			case Ncl_VECSUBSCR:
				for(size_t i=0; i<gsel->selection[lev_idx].u.vec.n_ind; i++){
					nlevs++;
				}
				break;
			default:
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Invalid subscripting method for dimension");
				throw pargal::pargal_except("Unable to create val slice, invalid subscripting method used", __FILE__, __LINE__, "Unknown");
		}
	}
	else{
		nlevs = 1;
	}
}

static void set_val_dims(std::vector<pargal::pcdim> &val_dim, const NclSelectionRecord *gsel)
{
	size_t n_elems = 0;
	assert(val_dim.size() == gsel->n_entries);
	for(size_t i=0; i<gsel->n_entries; i++){
		switch(gsel->selection[i].sel_type){
			case Ncl_SUB_ALL:
				// No change
				break;
			case Ncl_SUBSCR:
				n_elems = (gsel->selection[i].u.sub.finish - gsel->selection[i].u.sub.start)/(gsel->selection[i].u.sub.stride) + 1;
				val_dim[i].set_size(n_elems);
				break;
			case Ncl_VECSUBSCR:
				val_dim[i].set_size(gsel->selection[i].u.vec.n_ind);
				break;
			default:
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Invalid selector type");
		}
	}
}

static NclSelectionRecord *create_new_val_sel(const size_t  &ntsteps, const size_t &nlevs, std::vector<size_t> &new_val_dim_sz, NclSelectionRecord *old_sel)
{
	NclSelectionRecord *new_sel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
	if(!new_sel){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to allocate mem for new val selector");
		return NULL;
	}

	PNclSelCopy(new_sel, old_sel);

	size_t n_entries = new_sel->n_entries;

	new_val_dim_sz.resize(n_entries);
	PNclGetSelSz(new_sel, new_val_dim_sz);

	size_t lev_idx = n_entries - 3;
	size_t time_idx = lev_idx - 1;
	
	if(lev_idx >= 0){
		if(new_sel->selection[lev_idx].sel_type == Ncl_VECSUBSCR){
			NclFree(new_sel->selection[lev_idx].u.vec.ind);
		}
		new_sel->selection[lev_idx].sel_type = Ncl_SUB_ALL;
		new_sel->selection[lev_idx].u.sub.start = 0;
		new_sel->selection[lev_idx].u.sub.finish = nlevs - 1;
		new_sel->selection[lev_idx].u.sub.stride = 1;
		new_val_dim_sz[lev_idx] = nlevs;
	}

	if(time_idx >= 0){
		if(new_sel->selection[time_idx].sel_type == Ncl_VECSUBSCR){
			NclFree(new_sel->selection[time_idx].u.vec.ind);
		}
		new_sel->selection[time_idx].sel_type = Ncl_SUB_ALL;
		new_sel->selection[time_idx].u.sub.start = 0;
		new_sel->selection[time_idx].u.sub.finish = ntsteps - 1;
		new_sel->selection[lev_idx].u.sub.stride = 1;
		new_val_dim_sz[time_idx] = ntsteps;
	}

	return new_sel;
}

static void ParCALGetPcvarDims(pargal::pcvar *pcvar, std::vector<pargal::pcdim> &pcvar_dims)
{
	size_t ndims = pcvar->num_dims();
	pcvar_dims.resize(ndims);

	for(size_t i=0; i<ndims; i++){
		pcvar_dims[i] = pcvar->get_dim(i);
	}
}

ParCALMeshValSPtr ParCALMeshVal::create_mesh_val(NclSelectionRecord *from_gsel, NclSelectionRecord *from_lsel, const std::vector<size_t> &from_gdim_sz, const std::vector<size_t> &from_ldim_sz)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	MeshFInfo &mesh_finfo = mesh_->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle mb_set = pcal_finfo.get_file_set();
	moab::DataType dt = PNclNCMapToMB(data_type_);
	std::string pcvar_name = ParCALGetUniqStr(mesh_);
	const std::string &pcal_val_loc = pcvar_->get_loc();

	std::vector<int> to_val_tsteps;
	set_val_tsteps(pcvar_, from_gsel, to_val_tsteps);

	size_t to_nlevs = 0;
	set_val_nlevs(*this, from_gsel, to_nlevs);

	for(std::vector<int>::const_iterator iter = to_val_tsteps.begin(); iter != to_val_tsteps.end(); ++iter){
		size_t tstep = *iter;
		moab::EntityType ent_type = ParCALLocToMB(pcal_val_loc);
		mb_retval = PNclMBCreateTag(mesh_, pcvar_name.c_str(), tstep, to_nlevs, NULL, dt, ent_type);
		if(mb_retval != moab::MB_SUCCESS){
			std::string funcname("UNKNOWN");
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating MOAB tag while creating var");
			throw pargal::pargal_except(mb_core_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
		}
	}

//	std::vector<pargal::pcdim> to_val_dims = pcal_finfo.get_var_dims(pcvar_->name());
	std::vector<pargal::pcdim> to_val_dims;
	ParCALGetPcvarDims(pcvar_, to_val_dims);

	set_val_dims(to_val_dims, from_gsel); 
	mesh_->conform_mesh(to_val_dims, pcal_val_loc);
	pargal::pcvar *pcvar = new pargal::pcvar(pcvar_name, dt,
									to_val_dims, pcal_finfo, to_val_tsteps, true, pcal_val_loc);

	ParCALMeshVal *mesh_val = new ParCALMeshVal(pcvar, data_type_, false, missing_val_);
	mesh_val->num_pcvar_chunks_ = to_val_tsteps.size();

	std::vector<size_t> to_gdim_sz;
	NclSelectionRecord *to_gsel = create_new_val_sel(to_val_tsteps.size(), to_nlevs, to_gdim_sz, from_gsel);
	if(to_gsel == NULL){
		std::string err_msg("Error creating new val selector");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "%s", err_msg.c_str());
		throw pargal::pargal_except(err_msg, __FILE__, __LINE__, "UNKNOWN");
	}

	if(mesh_val->sel_ == NULL){
		// Only the first sel needs to be conformed to mesh, subsequent selectors
		// are selectors of a mesh_conformed val selector => already conforming
		mesh_->conform_mesh(to_gsel, pcal_val_loc);
	}
	mesh_val->sel_ = to_gsel;
	// FIXME: Find val loc from dim names
	size_t mb_chunk_sz = pcvar_->tstep_size();
	mesh_val->set_mesh_info(mesh_, mb_core_, mb_chunk_sz, ParCALMeshVal::MESHVAL_LOC_UNKNOWN);

	NclSelectionRecord *to_lsel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
	if(!to_lsel){
		std::string err_msg("Unable to alloc mem for new val local selector");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "%s", err_msg.c_str());
		throw pargal::pargal_except(err_msg, __FILE__, __LINE__, "UNKNOWN");
	}
	PNclSelCopy(to_lsel, to_gsel, 1);
	std::vector<size_t> to_ldim_sz(to_gdim_sz.begin() + 1, to_gdim_sz.end());

	//if(to_gdim_sz.size() == 4) printf("to gdim_sz before conv_to_local: %d, %d, %d, %d\n", to_gdim_sz[0], to_gdim_sz[1], to_gdim_sz[2], to_gdim_sz[3]);
	mesh_->conv_to_local(to_gsel, to_lsel, to_gdim_sz, to_ldim_sz, pcal_val_loc); 
	//if(to_ldim_sz.size() == 3) printf("to ldim_sz before conv_to_local: %d, %d, %d\n", to_ldim_sz[0], to_ldim_sz[1], to_ldim_sz[2]);

	/* Copy data to the new val */
	int new_val_tstep = 0;
	size_t to_buf_sz = pcvar->tstep_size();
	size_t from_buf_sz = pcvar_->tstep_size();
	for(std::vector<int>::const_iterator iter = to_val_tsteps.begin(); iter != to_val_tsteps.end(); ++iter){
		void *to_buf = pcvar->get_storage(new_val_tstep);
		void *from_buf = pcvar_->get_storage(*iter);

		double missing_val = 0.0;	
		switch(data_type_){
			case NC_FLOAT:
				missing_val = (missing_val_) ? ((double )*((float *)missing_val_)) : (((NclTypeClass)nclTypepdoubleClass)->type_class.default_mis.pdoubleval);
				break;
			case NC_DOUBLE:
				missing_val = (missing_val_) ? (*((double *)missing_val_)) : (((NclTypeClass)nclTypepdoubleClass)->type_class.default_mis.pdoubleval);
				break;
			default:
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Unknown data type");
				throw pargal::pargal_except("Unknown data type specified", __FILE__, __LINE__, "Unknown");
		}
		// FIXME: Don't assume double
		for(size_t i=0; i<to_buf_sz; i++){
			((double *)to_buf)[i] = missing_val;
		}
//		PNclSlideAndFill(ldim_sz, lsel, (double *)to_buf, to_buf_sz, (double *)from_buf, from_buf_sz);
		PNclSlideAndFill(to_ldim_sz, to_lsel, (double *)to_buf, to_buf_sz, from_ldim_sz, from_lsel, (double *)from_buf, from_buf_sz);

		if(is_fvar_){
			pcvar_->delete_storage(*iter);
		}
		new_val_tstep++;
	} 

	NclFree(to_lsel);
	
	return (ParCALMeshValSPtr(mesh_val));
}

ParCALMeshVal::ParCALMeshVal(ParCALMeshVal &val):prev_cached_lval_chunk_(parncl::PARCALVAL_NCHUNK_NONE), prev_cached_lval_buf_(NULL), ncached_lval_chunks_(0), mesh_(val.mesh_), mb_core_(val.mb_core_), val_loc_(ParCALMeshVal::MESHVAL_LOC_UNKNOWN), pcvar_(NULL), data_type_(val.data_type_), num_pcvar_chunks_(val.num_pcvar_chunks_), sel_(NULL), pcvar_chunk_sz_(val.pcvar_chunk_sz_), is_fvar_(false), ref_cnt_(0)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	parncl::MeshFInfo &mesh_finfo = mesh_->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	const std::vector<int>& pcvar_tsteps = val.pcvar_->get_tsteps();
	if(val.is_fvar_){
		// Record the root variable in the file
		root_fvar_name_ = val.root_fvar_name_;
	}
	std::string pcvar_name = ParCALGetUniqStr(mesh_);
	moab::DataType dt = PNclNCMapToMB(data_type_);
	moab::EntityType ent_type = ParCALLocToMB(val.pcvar_->get_loc());
	size_t nlevs = val.num_pcvar_levs_;

    for(std::vector<int>::const_iterator iter = pcvar_tsteps.begin(); iter != pcvar_tsteps.end(); ++iter){
        size_t to_tstep = *iter, from_tstep = *iter;

        val.get_cached_lval(*iter);
        if(val.is_fvar_){
            if(mesh_finfo.get_num_nc_files() > 1){
                /* If time steps are spread across files, 
                 * MOAB assumes that only 1 timestep exists in a file and 
                 * that loads any time step as timestep 0
                 */
                from_tstep = 0;
            }
        }
        mb_retval = PNclMBDupTag(mesh_, pcvar_name.c_str(), to_tstep, val.pcvar_->name().c_str(), from_tstep, nlevs, dt, ent_type);
        if(mb_retval != moab::MB_SUCCESS){
			// FIXME: Is there a portable way to get this info
			std::string funcname("UNKNOWN");
            NhlPError(NhlFATAL, NhlEUNKNOWN, "Error duplicating MOAB tag");
			throw pargal::pargal_except(mb_core_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
        }
		val.free_cached_lval(*iter);
    }
	// Create a new pcvar for the val
    const std::vector<pargal::pcdim>& pcvar_dims = val.pcvar_->get_dims();
    pcvar_ = new pargal::pcvar(pcvar_name, dt, pcvar_dims, pcal_finfo, pcvar_tsteps, true, val.pcvar_->get_loc());
	// FIXME: We don't catch memalloc failure
	cached_gval_ = std::make_pair(false, (std::vector<double> *)NULL);
}

ParCALMeshVal::~ParCALMeshVal()
{
	if(cached_gval_.first){
		delete cached_gval_.second;
	}
	if(pcvar_){
		delete pcvar_;
	}
	if(sel_){
		NclFree(sel_);
	}
}

ParCALMeshValSPtr ParCALMeshVal::slice(NclSelectionRecord *sel)
{
	// Convert sel to local sel - mesh func
	// gsel will be the selector stored in the new sliced val after conforming to mesh 
	NclSelectionRecord *gsel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
	if(gsel){
		PNclSelCopy(gsel, sel);
		if(sel_){
			gsel = PNclSelOfSel(sel_, gsel);
		}
	}
	else{
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to allocate memory for val sel");
		throw pargal::pargal_except("Mem alloc failure creating slice of var", __FILE__, __LINE__, "unknown");
	}

    const std::vector<pargal::pcdim>& pcvar_dims = pcvar_->get_dims();
	NclSelectionRecord *lsel = (NclSelectionRecord *) NclCalloc(1, sizeof(NclSelectionRecord));
	if(!lsel){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to allocate memory for temp local sel");
		throw pargal::pargal_except("Mem alloc failure creating slice of var", __FILE__, __LINE__, "unknown");
	}
	std::vector<size_t> gdims_sz(pcvar_dims.size(), 0);

	int i=0;
	for(std::vector<pargal::pcdim>::const_iterator iter = pcvar_dims.begin(); iter != pcvar_dims.end(); ++iter, i++){
		gdims_sz[i] = iter->size();
	}
	std::vector<size_t> ldims_sz(gdims_sz.begin() + 1, gdims_sz.end());
//	PNclGetSelSz(gsel, ldims_sz);
	assert(gdims_sz.size() == gsel->n_entries);
	for(int i=0; i<gsel->n_entries; i++){
		if(gsel->selection[i].sel_type == Ncl_SUBSCR){
			if((gsel->selection[i].u.sub.start == 0) && (gsel->selection[i].u.sub.finish == gdims_sz[i] - 1)){
				gsel->selection[i].sel_type = Ncl_SUB_ALL;
			}
		}
	}

	const std::string &pcal_val_loc = pcvar_->get_loc();
	//if(gdims_sz.size() == 4) printf("from gdim_sz before conv_to_local: %d, %d, %d, %d\n", gdims_sz[0], gdims_sz[1], gdims_sz[2], gdims_sz[3]);
	mesh_->conv_to_local(gsel, lsel, gdims_sz, ldims_sz, pcal_val_loc); 
	//if(ldims_sz.size() == 3) printf("from ldim_sz before conv_to_local: %d, %d, %d\n", ldims_sz[0], ldims_sz[1], ldims_sz[2]);

//	mesh_->conform_mesh(gsel, gdims_sz);

	// Copy tag using local selector
	ParCALMeshValSPtr sptr = create_mesh_val(gsel, lsel, gdims_sz, ldims_sz);

	// Only gsel is stored, lsel is temp
	NclFree(lsel);
	NclFree(gsel);

	return sptr;
}

void ParCALMeshVal::set_mesh_info(MBMesh *mesh, moab::Interface *mb_core, size_t chunk_sz, ParCALMeshVal::MeshValLoc val_loc)
{
	mesh_ = mesh;
	mb_core_ = mb_core;
	pcvar_chunk_sz_ = chunk_sz * num_pcvar_levs_;
	val_loc_ = val_loc;
}

ParCALMeshVal::MeshValLoc ParCALMeshVal::get_mesh_val_loc(void ) const
{
	return val_loc_;
}

pargal::fileinfo &ParCALMeshVal::get_pcal_finfo(void ) const
{
	MeshFInfo &mesh_finfo = mesh_->get_mesh_finfo();
	return mesh_finfo.get_pcal_finfo();
}

MBMesh *ParCALMeshVal::get_mesh(void ) const
{
	return mesh_;
}

size_t ParCALMeshVal::get_num_val_chunks(void ) const
{
	return num_pcvar_chunks_;
}

size_t ParCALMeshVal::get_num_val_levs(void ) const
{
	return num_pcvar_levs_;
}

size_t ParCALMeshVal::get_val_chunk_sz(void ) const
{
//	return pcvar_chunk_sz_;
	return pcvar_->tstep_size();
}

// FIXME: Don't assume double
std::vector<double> *ParCALMeshVal::get_gval(NclSelectionRecord *sel)
{
	if(cached_gval_.first){
		return cached_gval_.second;
	}

	if(cached_gval_.second){
		// FIXME: Although not optimal this is the easiest way
		delete cached_gval_.second;
		cached_gval_.second = NULL;
	}

	// FIXME: Don't assume double
	std::vector<double> *gdata = new std::vector<double>();
	gdata->clear();
	pargal::gather(*pcvar_, pcvar_->get_tsteps(), *gdata);

	if(gdata->size() > 0){
		const std::vector<pargal::pcdim>& var_pcdims = pcvar_->get_dims();
		std::vector<size_t> var_dims(var_pcdims.size(), 0);
		int i=0;
		for(std::vector<pargal::pcdim>::const_iterator iter = var_pcdims.begin(); iter!= var_pcdims.end(); ++iter, i++){
			var_dims[i] = iter->size();
		}
		if(sel && sel_){
			sel_ = PNclSelOfSel(sel, sel_);
		}

		PNclSlideAndFill(var_dims, (sel_)?(sel_):(sel), (double *)&((*gdata)[0]), gdata->size());
	}

	cached_gval_.second = gdata;
	cached_gval_.first = true;

	return cached_gval_.second;
}

void *ParCALMeshVal::get_cached_lval(size_t chunk)
{
	void *buf = NULL;
	// FIXME: Make MB design more flexible to get rid of this ugly assert
	if(ncached_lval_chunks_ == 0){
		buf = (void *)pcvar_->get_storage(chunk);
		prev_cached_lval_chunk_ = chunk;
		prev_cached_lval_buf_ = buf;
	}
	else{
		assert(chunk == prev_cached_lval_chunk_);
		buf = prev_cached_lval_buf_;
	}
	ncached_lval_chunks_++;
 	return buf;
}

void ParCALMeshVal::free_cached_lval(size_t chunk)
{
	ncached_lval_chunks_--;
	assert(chunk == prev_cached_lval_chunk_);
	if(is_fvar_ && (ncached_lval_chunks_ == 0)){
		pcvar_->delete_storage(prev_cached_lval_chunk_);	
		prev_cached_lval_chunk_ = parncl::PARCALVAL_NCHUNK_NONE;
		prev_cached_lval_buf_ = NULL;
	}
}

pargal::pcvar *ParCALMeshVal::get_pcvar(void )
{
	return pcvar_;
}

nc_type ParCALMeshVal::get_data_type(void ) const
{
	return data_type_;
}

void *ParCALMeshVal::get_missing_val(void ) const
{
	return missing_val_;
}

const NclSelectionRecord  *ParCALMeshVal::get_val_sel(void ) const
{
	return sel_;
}

bool ParCALMeshVal::is_file_val(void ) const
{
	return is_fvar_;
}

int ParCALMeshVal::get_ref_cnt(void ) const
{
	return ref_cnt_;
}

void ParCALMeshVal::inc_ref_cnt(void )
{
	ref_cnt_++;
}

void ParCALMeshVal::dec_ref_cnt(void)
{
	ref_cnt_--;
}

void ParCALMeshVal::set_gval_cache_dirty(void )
{
	cached_gval_.first = false;
}

} // namespace parncl
