/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef ParCALMeshVal_hpp
#define ParCALMeshVal_hpp

#include <vector>
#include <utility>
#include "ParCALCore.hpp"
#include "moab/Core.hpp"
#include "moab/Types.hpp"
extern "C" {
#include "ParCALFuncs.h"
}

namespace parncl{

	class ParCALMeshVal;
	/* FIXME: Do we need to make this sptr more generic? */
	class ParCALMeshValSPtr{
		public:
			ParCALMeshValSPtr(const ParCALMeshValSPtr &sptr);
			ParCALMeshValSPtr(ParCALMeshVal *valp);

			ParCALMeshVal *operator->();
			ParCALMeshVal &operator*();	
			ParCALMeshValSPtr& operator=(const ParCALMeshValSPtr &rhs);

			~ParCALMeshValSPtr();
			
		private:
			ParCALMeshValSPtr();

			ParCALMeshVal *valp_;

			friend class ParCALMeshVal;
	};

	const int PARCALVAL_NCHUNK_NONE = -1;

	class ParCALMeshVal{
		public:
			enum MeshValLoc{
				MESHVAL_ON_VERTEX,
				MESHVAL_ON_QUAD_CENTER,
				MESHVAL_ON_LAT_EDGE,
				MESHVAL_ON_LON_EDGE,
				MESHVAL_LOC_UNKNOWN
			};

			static std::string convert_to_parcal_loc(const MeshValLoc val_loc);

			// FIXME: For create_mesh_val() we cud add the mesh instance since it is
			// clear where this val will go to...
			// create mesh val corresponding to an existing pcvar
			static ParCALMeshValSPtr create_mesh_val(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar, void *missing_val);
			// create mesh val with the given dimensions on the given mesh
			static ParCALMeshValSPtr create_mesh_val(MBMesh *mesh, moab::Interface *mb_core, size_t mb_chunk_sz, const std::vector<int> &val_tsteps, const std::vector<pargal::pcdim> &val_dims, size_t val_nlevs, nc_type val_dt, const std::string &pcal_val_loc, void *missing_val, NclSelectionRecord *sel); 

			// Get a slice of a val
			ParCALMeshValSPtr slice(NclSelectionRecord *sel);

			// Get no of val chunks
			size_t get_num_val_chunks(void ) const;
			// Get no of levs in the val
			size_t get_num_val_levs(void ) const;
			// Get chunk sz
			size_t get_val_chunk_sz(void ) const;
			// Set mesh info - Once the variable is placed on a mesh
			// mesh related info should be set
			// Once mesh classes come in it shud just have a mesh instance as arg
			void set_mesh_info(MBMesh *mesh, moab::Interface *mb_core, size_t mb_chunk_sz, MeshValLoc val_loc_on_mesh);

			/* Get Mesh val loc */
			// FIXME: Do we need to expose the mesh val loc?
			MeshValLoc get_mesh_val_loc(void ) const;

			/* Get parcal core hnd corresponding to this var */
			pargal::fileinfo &get_pcal_finfo(void ) const;
			MBMesh *get_mesh(void ) const;
			/* Get global val - gvals are mutable */
			// FIXME: Don't assume double
			std::vector<double> *get_gval(NclSelectionRecord *sel=NULL);
			/* Get lval & cache it - when used as a rvalue */
			void *get_cached_lval(size_t chunk);
			/* Free the cached lval */
			void free_cached_lval(size_t chunk);
			/* Get pcvar corresponding to this mesh val */
			pargal::pcvar *get_pcvar(void );
			/* Get NetCDF type of the data stored by the mesh val */
			nc_type get_data_type(void ) const;
			/* Get missing val */
			void *get_missing_val(void ) const;
			/* Get val selector */
			const NclSelectionRecord *get_val_sel(void ) const;
			/* Check if the mesh val was read from a file or not */
			bool is_file_val(void ) const;

		private:
			ParCALMeshVal();
			~ParCALMeshVal();
			ParCALMeshVal(pargal::pcvar *pcvar, nc_type data_type, bool is_fvar, void *missing_val);
			ParCALMeshVal(ParCALMeshVal &val);

			ParCALMeshValSPtr create_mesh_val(NclSelectionRecord *gsel, NclSelectionRecord *lsel, const std::vector<size_t> &gdim_sz, const std::vector<size_t> &ldim_sz);

			/* Increment/decrement val ref counts */
			int get_ref_cnt(void ) const;
			void inc_ref_cnt(void );
			void dec_ref_cnt(void );

			/* Set gval cache is dirty */
			void set_gval_cache_dirty(void );

			/* Number of refs to this val */
			/* FIXME: Multiple sptrs will modify this ref_cnt, we are not MT yet so... */
			int ref_cnt_;
			/* 
			 * cached_lval_chunk_no = index in this vector
			 * std::pair<ref_cnt_for_chunk, cached_lval_chunk_bufp>
			 * FIXME: The current MOAB NC reader design does not allow us to
			 * cache multiple lvals
			 */
			//std::vector<std::pair<int, void *> > cached_lvals_;
			/* The current MOAB design forces us to use this potentially
			 * dangerous way of caching/deleting lvals
			 */
			size_t prev_cached_lval_chunk_;
			void *prev_cached_lval_buf_;
			size_t ncached_lval_chunks_;
			/* <cached_gval_is_valid_flag, cached_gval_bufp> */
			// FIXME: Don't assume double
			std::pair<bool, std::vector<double>*> cached_gval_;

			// FIXME: Some of the vars below shud be replaced by mesh instance
			MBMesh *mesh_;
			moab::Interface *mb_core_;
			MeshValLoc val_loc_;
			/* ParCAL variable corresponding to this val */
			pargal::pcvar *pcvar_;
			const nc_type data_type_;
			void *missing_val_;
			size_t num_pcvar_chunks_;
			size_t pcvar_chunk_sz_;
			size_t num_pcvar_levs_;
			/* This sel_ is NOT set via a VarSHnd, it is set internally within 
			 * a MeshVal. Since ParGAL does not support sparse tags yet, we
			 * need to store some slice info here.
			 */
			NclSelectionRecord *sel_;
			/* If the val corresponds to a file var */
			const bool is_fvar_;
			std::string root_fvar_name_;

			friend class ParCALMeshValSPtr;
	};
}

#endif // ParCALMeshVal_hpp
