/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "ParCALVarSHnd.hpp"
#include "PNclMDUtils.hpp"
#include <iostream>

namespace parncl{

ParCALVarSHnd::ParCALVarSHnd(ParCALMeshValSPtr &sptr):sptr_(sptr), sel_(NULL),declone_caller_level_(-1)
{
}

ParCALVarSHnd::ParCALVarSHnd(ParCALMeshValSPtr &sptr, NclSelectionRecord *sel):sptr_(sptr), sel_(sel), declone_caller_level_(-1)
{
}

ParCALVarSHnd::ParCALVarSHnd(const ParCALVarSHnd &var):sptr_(var.sptr_), declone_caller_level_(-1)
{
	sel_ = (NclSelectionRecord *)NclMalloc(sizeof(NclSelectionRecord));
	if(sel_ && var.sel_){
		PNclSelCopy(sel_, var.sel_);	
	}
}

ParCALVarSHnd::ParCALVarSHnd(const ParCALVarSHnd &var, NclSelectionRecord *sel):sptr_(var.sptr_), sel_(NULL), declone_caller_level_(-1) 
{
	if(var.sel_){
		NclSelectionRecord *mysel = NULL;
		mysel = (NclSelectionRecord *)NclMalloc(sizeof(NclSelectionRecord));
		if(mysel){
			PNclSelCopy(mysel, var.sel_);
		}
		else{
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to allocate memory for temp sel");
			throw pargal::pargal_except("Mem alloc failure creating var", __FILE__, __LINE__, "unknown");
		}
		mysel = PNclSelOfSel(var.sel_, mysel);
		// slice a var
		sptr_ = sptr_->slice(mysel);
		NclFree(mysel);
	}
	else{
		// slice a var
		sptr_ = sptr_->slice(sel);
	}
}


ParCALVarSHnd::~ParCALVarSHnd()
{
	if(sel_){
		free(sel_);
	}
}

int ParCALVarSHnd::get_num_val_chunks(void )
{
	return sptr_->get_num_val_chunks();
}

// FIXME: val_type and req_chunk convey the same info... atleast for now
void *ParCALVarSHnd::get_val(NclSelectionRecord *sel, size_t req_chunk, ParCALValType val_type, ParCALAccessMode mode, size_t &sz)
{
	void *buf = NULL;
	sz = 0;
	
	// FIXME : Allow getting a selection of data
	assert(sel == NULL);
	// Global val is never on the mesh
	if((mode & PARCAL_WR) && (val_type != PARCAL_VAL_GLOBAL)){
		// Copy
		sptr_ = ParCALMeshValSPtr(&(*sptr_));
	}
	if(val_type == PARCAL_VAL_LOCAL){
		buf = sptr_->get_cached_lval(req_chunk);
		sz = sptr_->get_val_chunk_sz();
	}
	else{
		// FIXME: Don't assume doubles
		std::vector<double> *gdata= sptr_->get_gval(sel_);
		if(gdata){
			buf = (void *)&((*gdata)[0]);
			sz = gdata->size();
		}
	}
	return buf;
}

void ParCALVarSHnd::delete_val(int req_chunk)
{
	// Don't need to invalidate gvals
	if(req_chunk != parncl::PARCALVAR_NCHUNK_ALL){
		sptr_->free_cached_lval(req_chunk);
	}
}

pargal::pcvar *ParCALVarSHnd::get_pcvar(void )
{
	return sptr_->get_pcvar();
}

size_t ParCALVarSHnd::get_num_val_levs(void )
{
	return sptr_->get_num_val_levs();
}

pargal::fileinfo &ParCALVarSHnd::get_pcal_finfo(void )
{
	return sptr_->get_pcal_finfo();
}

MBMesh *ParCALVarSHnd::get_mesh(void )
{
	return sptr_->get_mesh();
}

nc_type ParCALVarSHnd::get_data_type(void )
{
	return sptr_->get_data_type();
}

void *ParCALVarSHnd::get_missing_val(void )
{
	return sptr_->get_missing_val();
}

ParCALMeshVal::MeshValLoc ParCALVarSHnd::get_mesh_val_loc(void )
{
	return sptr_->get_mesh_val_loc();
}

void ParCALVarSHnd::get_sel(NclSelectionRecord *sel)
{
	const NclSelectionRecord *val_sel = sptr_->get_val_sel();
	if(sel_){
		PNclSelCopy(sel, sel_);
		if(val_sel){
			sel = PNclSelOfSel(val_sel, sel);
		}
	}
	else{
		if(val_sel){
			PNclSelCopy(sel, val_sel);
		}
		else{
			sel->n_entries = 0;
		}
	}
}

void ParCALVarSHnd::set_declone_caller_level(int caller_level)
{
	declone_caller_level_ = caller_level;
}

int ParCALVarSHnd::get_declone_caller_level(void ) const
{
	return declone_caller_level_;
}

void ParCALVarSHnd::declone(void )
{
	declone_caller_level_ = PARNCL_INVALID_CALLER_LEVEL;
}

} // namespace parncl
